
from ..scaffolds import ScaffoldRing, BindingMode
from ..extraframework import Metal, Ligand, Combination, hydroxide, oxide, hydride
from ..clusterizer import *
from ..gaussian import G09Output
import os
import pprint


scaffold = ScaffoldRing(r'/projects/bioenergy/ngallup/git/zeoliteclusterizer/zeoliteclusterizer/tests/optimized_ring/optimized_ring.com')

print scaffold.getLoc()
print scaffold.charge
print scaffold.unpaired
print scaffold.atom_list

modes_dir = os.path.abspath(r'/projects/bioenergy/ngallup/git/zeoliteclusterizer/zeoliteclusterizer/tests/binding_sites')
modes_coms = os.listdir(modes_dir)
modes_coms = [os.path.join(modes_dir, each) for each in modes_coms]

# Note: create a BindingModes container class that takes a directory and
# recurses all .com and .gjf files
modes_coms = [BindingMode(each, scaffold.atom_list) for each in modes_coms]
pprint.pprint([each.metal_ligand_list for each in modes_coms])

# Note: might be worth considering a Metals container class for all Metal
# objects to be easier to read and keep track of.  Might be overengineering
# but is probably more intuitive than requring the user to submit a list of
# objects, especially when only concerned with one metal
#zn = Metal('Zn', [2], [0])
#cu = Metal('Cu', [1], [0])
#cr = Metal('Cr', [3], [0])
sc = Metal('Sc', [1,3], [0])
ti = Metal('Ti', [2,4], [0])
v = Metal('V', [1,3,5], [0])
cr = Metal('Cr', [2,4,6], [0])
mn = Metal('Mn', [3,5], [0])
fe = Metal('Fe', [2], [0])
co = Metal('Co', [1,3], [0])
ni = Metal('Ni', [2], [0])
cu = Metal('Cu', [2], [0])
zn = Metal('Zn', [2], [0])
metals = [sc, ti, v, cr, mn, fe, co, ni, cu, zn]

# Consider a Ligands container class for easier reading
print hydroxide.name, hydroxide.atom_list
print hydride.name, hydride.atom_list
print oxide.name, oxide.atom_list
ligands = [hydroxide, hydride, oxide]

#test = Combination()
#test.addMetal(zn, 2, 0)
#test.addLigand(hydroxide, -1, 0)
#print 'strings: ', test._strings
#print 'Hash: ', test._hashval
#print 'Metal: ', test.metal
#print 'Ligands: ', test.ligands
clusters = Clusterizer(scaffold, modes_coms, metals, 
					ligands, charges=[0], unpaired=[[0]], 
					mix_ligands=False)

output = G09Output(r'/projects/bioenergy/ngallup/git/zeoliteclusterizer/zeoliteclusterizer/tests/tests')
output.writeAllModes(clusters.getFinalModes(), makedirs=True)




from abc import ABCMeta, abstractmethod
import string
import os

class GaussianInput(object):
	'''
	An abstract class containing useful base methods for handling
	and parsing Gaussian input files (Specifically G09-type)
	'''
	
	__metaclass__ = ABCMeta

	@abstractmethod
	def getLoc():
		'''Return absolute file location'''
		pass

	def readHeader(self, line_list):
		# Need to read in header information and parse out extraneous stuff
		# (chk, etc.)

		header = []
		title_index = 0
		charge = 0
		mult = 0
		unpaired_e = 0

		for index, line in enumerate(line_list):
			if 'Title Card Required' in line:
				title_index = index
				break
			if '%' in line:
				continue
			header.append(line.rstrip())

		# Need to get rid of extra line between head and title card
		header.pop()
		header = str(*header)

		# Need to gather charges and unpaired electrons
		charge_mult = line_list[title_index+2].split()
		charge = int(charge_mult[0])
		mult = int(charge_mult[1])
		unpaired_e = mult - 1

		atom_start = title_index + 3

		return header, charge, unpaired_e, atom_start

	def readAtoms(self, start_ind, lines):
		# Returns a list of atomtype, frozen(y/n), x, y, z coordinates

		atom_list = []

		for index, line in enumerate(lines[start_ind:]):
			atomtype = None
			frozen = None
			x, y, z = None, None, None

			split_line = list(line.split())

			if self.startsWithLetter(split_line) == False:
				break
			elif self.frozenIsSpecified(split_line) == False:
				self.addSplit(split_line, atom_list, frozen='0')
			else:
				self.addSplit(split_line, atom_list)

		return atom_list
	
	def addSplit(self, split_line, atom_list, frozen=None):
		# Helper function for splitting atom lines into components.  If frozen
		# is default, parses like normal.  Otherwise utilizes the provided
		# value for frozen

		atomtype = None
		x, y, z = None, None, None

		if frozen == None:
			atom_list.append(split_line)
		else:
			atomtype = split_line[0]
			x, y, z = split_line[1], split_line[2], split_line[3]
			result = [atomtype, frozen, x, y, z]
			atom_list.append(result)

	def frozenIsSpecified(self, atom_line):
		# Returns boolean of whether or not a frozen/non-frozen coordinate
		# is specified within a gaussian input line

		if atom_line[1] == '0' or atom_line[1] == '-1':
			return True
		return False

	def startsWithLetter(self, atom_line):
		# Return boolean of whether or not the provided input line begins
		# with a letter, primarily for identifying atom lines

		if atom_line == []:
			return False
		elif atom_line[0][0] in string.ascii_letters:
			return True
		return False

class OutputTest(object):
	def write(self, num, mode):
		with open('tests/test%s.com' % num, 'w') as outfile:
			for line in mode.scaffold.head_lines:
				outfile.write(''.join(line))
			outfile.write('\n\nTest\n\n0 1\n')
			for line in mode.scaffold.atom_list:
				outfile.write(' '.join(line))
				outfile.write('\n')
			for line in mode.conformations:
				outfile.write(' '.join(line))
				outfile.write('\n')

class Output(object):
	'''
	Abstract base class for all output file formats
	'''

	__metaclass__ = ABCMeta

	@abstractmethod
	def writeAllModes():
		pass
	@abstractmethod
	def write():
		pass

	def __init__(self, disp_frozen, atom_width, frozen_width, 
					coord_width, coord_digits):
		self.frozen = disp_frozen
		self.atom_width = atom_width
		self.frozen_width = frozen_width
		self.coord_width = coord_width
		self.coord_digits = coord_digits

class G09Output(Output):
	'''
	For writing final binding modes to disk in a G09 compatible fashion
	'''

	def __init__(self, dir):
		super(G09Output, self).__init__(True, 5, 5, 7, 5)
		self.dir = os.path.abspath(dir)

	def writeAllModes(self, absModes, header=None, footer=None, makedirs=False):
		'''
		Write all conformations for the provided AbstractModes into the provided
		directory.  Use hash table to iterate conformation number.  See write
		method for more details.
		'''
		# Note: writeAllModes is not properly set up to handle the mixed
		# conformations that would exists as mixed ligand sets in 
		# mode.conformations.
		hashtable = {}

		for mode in absModes:
			hval = mode.combo._hashval
			if hval not in hashtable:
				hashtable[hval] = 0
			hashtable[hval] += 1
			conf = hashtable[hval]

			name = [s for s in mode.combo._strings]
			charge = str(mode.combo.charge + mode.scaffold.charge)
			unpaired = (mode.combo.unpaired + mode.scaffold.unpaired) % 2
			mult = str(unpaired+1)
			unpaired = str(unpaired)
			name.append('_charge%s_%set_conf%d' % (charge, mult, conf))
			name = ''.join(name)
			
			name = os.path.join(self.dir, name)
			if makedirs == True:
				os.makedirs(name)
				name = os.path.join(name, os.path.basename(name))

			self.write(mode, name, charge=charge, mult=mult)
	
	def write(self, mode, name, charge='0', mult='1', header=None, footer=None):
		'''
		Write a single file.  Name should have the absolute path prepended to the
		.com filename.  If alternative headers and footers are provided, write
		those instead of what's provided in the ScaffoldRing.  The header should
		exclude the Title Card Required line because it will be added in by the
		method.
		'''
		if header == None:
			header = mode.scaffold.head_lines

		with open(name + '.com', 'w') as comfile:
			check = '%chk=' + os.path.basename(name) + '.chk' + '\n'
			comfile.write(check)
			comfile.write(header.rstrip('\n') + '\n')
			
			comfile.write('\nAutomatically Generated by ZeoliteClusterizer\n\n')
			comfile.write('%s %s\n' % (charge, mult))

			# Refactor these for loops as some point
			for line in mode.scaffold.atom_list:
				# Think of a way to make this not look like garbage
				params = (self.atom_width, line[0], self.frozen_width, line[1], 
							self.coord_digits, float(line[2]), self.coord_digits, 
							float(line[3]), self.coord_digits, float(line[4]))
				comfile.write('%-*s %-*s %-*f %-*f %-*f\n' % params)

			for line in mode.conformations:
				params = (self.atom_width, line[0], self.frozen_width, line[1],
                     self.coord_digits, float(line[2]), self.coord_digits,
                     float(line[3]), self.coord_digits, float(line[4]))
				comfile.write('%-*s %-*s %-*f %-*f %-*f\n' % params)

			comfile.write('\n')
			if footer != None:
				comfile.write(footer)

			comfile.write('\n\n\n')




import sys
from extraframework import Combination
from scaffolds import AbstractMode
from gaussian import OutputTest

class Clusterizer(object):
	'''
	The Clusterizer is the meat of ZeoliteClusterizer, hence the name.  It tries
	to take advantage of dynamic programming practices to streamline the
	metal-ligand combinatorics.

	The basic intended workflow is like this:
	(1) Place binding modes in hash table, using ligand number as key
	(2) Feed parameters to combinations method, accumulate allowed Combinations
			using DP concepts based on electronic parameters
	(3) Iterate over allowed combinations, matching to modes in hash table
	(4) Create conformational variety of each Combination-Mode match
	(5) Write conformations to disk
	'''
	def __init__(self, scaffold, binding_list, metals, 
					ligands, charges=[0], unpaired=[[0]], mix_ligands=True):

		# Need hash tables of binding modes and counter charges
		self.maxLigs = 0
		self._modes = self.makeModesHashTable(binding_list)
		self._cCharges = self.makeChargesHashTable(scaffold, charges, unpaired)
		self.combinations = []
		self.finalModes = None

		if mix_ligands == False:
			for metal in metals:
				self.makePureCombinations(scaffold, binding_list, metal, ligands, 
													charges, unpaired)
		else:
			self.makeMixCombinations(scaffold, binding_list, metal, ligands,
											charges, unpaired, combos)

		abstractModes = [AbstractMode(combo, mode, scaffold, mixed=False) 
			for combo, mode in self.matchingModes(self.combinations, self._modes)]

		self.finalModes = abstractModes

	def makeModesHashTable(self, binding_modes):
		'''
		Returns a hash table consisting of binding modes with the ligand number
		as the key with values consisting of a list of modes that contain that
		number of ligands
		'''
		hashTable = {}

		for mode in binding_modes:
			ligNum = mode.getNumLigands()
			if ligNum > self.maxLigs:
				self.maxLigs = ligNum
			if ligNum not in hashTable:
				hashTable[ligNum] = [mode]
			else:
				hashTable[ligNum].append(mode)
		
		return hashTable

	def makeChargesHashTable(self, scaffold, charges, unpaireds):
		hashTable = {}
		for charge, unpaired in zip(charges, unpaireds):
			charge -= scaffold.charge
			hashTable[charge] = unpaired
		return hashTable

	def makePureCombinations(self, scaffold, binding_list, metal, ligands, 
									charge, unpaired):
		'''
		Since we're only specifying pure ligand combinations, we can use
		simple math to construct combinations
		'''
		prevCombos = {}

		# This needs to be refactored ASAP because it reads like ass.  Maybe
		# implement a method object or some other abstract concept.  Maybe a
		# generator would be a good fit.  Maybe just break into smaller methods
		# Note: this is not yet handling counter radicals like it should in the
		# _cCharge dict
		for metalCharge in metal.charge:
			for metalUnpaired in metal.unpaired:
				for lig in ligands:
					newCombo = Combination()
					newCombo.addMetal(metal, metalCharge, metalUnpaired)
					if not self.isPrevCombo(newCombo, prevCombos):
						if (self.isCounterCharge(newCombo) and
								newCombo.getNumLigands() in self._modes):
							self.combinations.append(newCombo)
					prevCombos = self.addToComboDict(newCombo, prevCombos)
					for charge in lig.charge:
						for cCharge in self._cCharges:
							tmpCombo = None
							ligsToAdd = (newCombo.charge - cCharge) / (-charge)
							if (ligsToAdd > 0 and ligsToAdd in self._modes):
								tmpCombo = self.addLigs(newCombo, lig, ligsToAdd)
								if not self.isPrevCombo(tmpCombo, prevCombos):
									self.combinations.append(tmpCombo)
									prevCombos = self.addToComboDict(tmpCombo, 
																				prevCombos)

	
	def isCounterCharge(self, combination):
		if combination.charge in self._cCharges:
			return True
		return False

	def isPrevCombo(self, combination, combo_dict):
		if combination._hashval in combo_dict:
			return True
		return False

	def addToComboDict(self, combination, combo_dict):
		combo_dict[combination._hashval] = combination
		return combo_dict

	def addLigs(self, combination, ligand, numToAdd):
		for _ in range(numToAdd):
			# Note: change ligand class to only have one charge and one 
			# unpaired.  Otherwise is overcomplicated
			charge = list(ligand.charge)[0]
			unpaired = list(ligand.unpaired)[0]
			combination.addLigand(ligand, charge, unpaired)
		return combination

	def makeMixCombinations(self, scaffold, binding_list, metal, ligands,
									charge, unpaired):
		print("Mixed combinations not implemented yet!")
		sys.exit()

	def matchingModes(self, combo_list, modes_list):
		for combo in combo_list:
			numLigs = combo.getNumLigands()
			for mode in modes_list[numLigs]:
				yield combo, mode

	def getFinalModes(self):
		return self.finalModes




import sys
import string

class Properties(object):
	'''
	A base class for ligands, metals, etc., that contains some basic
	chemical properties.  Charges and number of unpaired electrons can
	be provided as a list if multiple states are phyiscally accessible, as is
	often the case with metals
	'''
	def __init__(self, charge_list, unpaired_list):
		try:
			self.charge = set(charge_list)
			unpaired_list = [each % 2 for each in unpaired_list]
			self.unpaired = set(unpaired_list)
		except:
			print("You probably gave me an int when it should have been an iterable",
					sys.exc_info()[0])
			raise


class Metal(Properties):
	'''
	A class for metals to contain physical properties.
	'''
	def __init__(self, name, charges, unpaired):
		super(Metal, self).__init__(charges, unpaired)
		self.name = str(name)

class Ligand(Metal):
	'''
	A class for handling ligands which can consist of more than one atom and
	could potentially have multiple charge and multiplicity states, although
	this latter possibility would be rare.

	atom_list should consist of a string and three numbers, e.g.:
		[['H', 0.0, 0.0, 1.0], ...]
	'''
	def __init__(self, name, atom_list, charges, unpaired, 
						headInd=0, tailInd=None):
		super(Ligand, self).__init__(name, charges, unpaired)

		self._head, self._headIndex = None, None
		self._tail, self._tailIndex = None, None

		# Will defer final print to Gaussian output handler
		self.atom_list = self.convertToFloats(atom_list)

		# Need to set the connecting atom; can be changed by user
		self.setHead(headInd)
		if tailInd == None:
			if len(self.atom_list) > 1:
				self.setTail(1)
			else:
				self.setTail(0)
		else:
			self.setTail(tailInd)

	def convertToFloats(self, atom_list):
		'''
		It is conceivable that a user might read in coordinates from a file,
		thus it's important that the resulting strings be converted into floats
		to maintain the ability to easily transform coordinates, and control
		final output printing
		'''
		newList = []
		for element in atom_list:
			# Want to ignore elemental string and examine rest
			tmp = [element[0]]
			for i in range(1, 4):
				element[i] = float(element[i])
				tmp.append(element[i])
			newList.append(tmp)

		return newList

	def setHead(self, index):
		'''
		The head atom is the atom connecting directly to the metal.  It is the
		gateway into the rest of the ligand.  The head and tail are different
		only when the ligand is greater than 1 atom
		'''
		self._head = self.atom_list[index]
		self._headIndex = index
	
	def setTail(self, index):
		'''
		The tail is the second atom used to form a vector between the head and
		the tail that is utilized for geometric transformations
		'''
		self._tail = self.atom_list[index]
		self._tailIndex = index

	def getHead(self):
		return self._head

	def getTail(self):
		return self._tail

	def getHeadIndex(self):
		return self._headIndex

	def getTailIndex(self):
		return self._tailIndex 

class Combination(object):
	'''
	'''
	def __init__(self):
		self.metal = None
		self.ligands = []
		self._hashval = None
		self._strings = []
		self.charge = 0
		self.unpaired = 0

	def isMetal(self, metal):
		if type(metal) == Metal:
			return True
		return False

	def isLigand(self, lig):
		if type(lig) == Ligand:
			return True
		return False

	def addMetal(self, metal, charge, unpaired):
		self.metal = metal
		self._strings.append(metal.name)
		self.updateHashVal(charge, unpaired)

	def addLigand(self, lig, charge, unpaired):
		self.ligands.append(lig)
		self._strings.append(lig.name)
		self.updateHashVal(charge, unpaired)

	def updateHashVal(self, charge, unpaired):
		# Implement binary insertion at some point or implement custom char
		# order ignorant hashing scheme (e.g. h = sum(c*x^2)) to avoid the
		# O(n log n) sort and O(n) join
		self.updateElectrons(charge, unpaired)
		self._strings = sorted(self._strings)
		tmp = [ string for string in self._strings ]
		tmp.append( '(%d)(%d)' % (self.charge, self.unpaired))
		self._hashval = ''.join(tmp)
		#self._hashval = self._hashval.join(tmp)

	def updateElectrons(self, charge, unpaired):
		self.charge += charge
		self.unpaired = (self.unpaired + unpaired) % 2

	def getNumLigands(self):
		return len(self.ligands)
	
	
'''
Some common ligands are listed below
'''
hydroxide_pos = [
            ['O', 0.00000000, 0.00000000, 0.00000000],
            ['H', 0.00000000, 0.00000000, -1.00000000]
         ]
hydroxide = Ligand('OH', hydroxide_pos, [-1], [0])

hydride_pos = [ ['H', 0.0, 0.0, 0.0] ]
hydride = Ligand('Hydride', hydride_pos, [-1], [0])

oxide_pos = [ ['O', 0.0, 0.0, 0.0] ]
oxide = Ligand('Oxide', oxide_pos, [-2], [0])



import string
from gaussian import GaussianInput
import os, sys
import numpy as np
import copy

class ScaffoldRing(GaussianInput):
	'''
	For defining the zeolite scaffold ring and its inherent chemical and
	electronic properties, as well as providing an easy collection of atom
	coordinates and characteristics
	'''
	def getLoc(self):
		return self.input_loc

	def __init__(self, comfile):
		self.input_loc = os.path.abspath(comfile)
		with open(comfile,'r') as infile:
			lines = infile.readlines()
			
			self.head_lines, self.charge, self.unpaired, self.atom_start = self.readHeader(lines)
			self.atom_list = self.readAtoms(self.atom_start, lines)

class BindingMode(GaussianInput):
	'''
	For defining the binding mode rings to be permutated upon.  Primarily for
	parsing out the metal binding modes and ligand locations.
	'''
	def __init__(self, comfile, scaffold_atoms):
		self.input_loc = os.path.abspath(comfile)
		with open(comfile, 'r') as infile:
			lines = infile.readlines()

			self.metal_ligand_list = self.getMetalsLigands(lines, scaffold_atoms)

	def getMetalsLigands(self, lines, scaffold_atoms):
		'''
		Returns list of metal and ligand locations.  Current implementation
		is very primative: the method simply compares the list lengths and
		truncates the leading atoms.  More advanced handling should be
		implemented at a later time.
		'''
		atom_list = []
		scaffold_len = len(scaffold_atoms)

		# Need to find the beginning of the atom list
		index = 0
		while('Title Card Required' not in lines[index]):
			index += 1
		index += 3

		# Want to truncate atom list
		index += scaffold_len
		line = lines[index].split()
		while(self.startsWithLetter(line) == True):
			self.addSplit(line, atom_list)
			index += 1
			line = lines[index].split()

		return atom_list
	
	def getLoc(self):
		return self.input_loc

	def getNumLigands(self):
		'''
		Currently assumes only a single metal.  Should be good approximation
		but one day may need to be changed.
		'''
		return len(self.metal_ligand_list)-1

	def getMetal(self):
		'''
		Specifically looks for "X" as the metal to be replaced
		'''
		result = []
		for line in self.metal_ligand_list:
			if line[0] == 'X':
				result.append(line)
		return result

	def getLigands(self):
		result = []
		for line in self.metal_ligand_list:
			if line[0] != 'X':
				result.append(line)
		return result

class AbstractMode(object):
	'''
	A class to bind a particular Combination and BindingMode object to make it
	easier to iterate upon all possible conformations and finally write to an
	output file
	'''
	def __init__(self, combo_obj, binding_obj, scaffold, mixed=True):
		self.combo = combo_obj
		self.bindingMode = binding_obj
		self.scaffold = scaffold
		self.conformations = self._createConformers(self.combo, 
													self.bindingMode, mixed)

	def _createConformers(self, combo, binding, mixed):
		'''
		Create all the various possible conformations.  Dynammic programming is
		overkill but a good practice.  If ligands aren't mixed, don't need to
		actually create conformational variety because of symmetry.
		'''
		conformations = []
		prevConfs = {}

		initial = copy.deepcopy(binding.metal_ligand_list) # Create copy
		initial = self._replaceMetal(combo, initial)

		ligandList = combo.ligands

		if mixed:
			# Remember to make use of the prevConfs dict for DP, like would be
			# necessary for 2OH's and a hydride
			conformations = self._transformMixedLigs()
		else:
			conformations = self._transformSameLigs(initial, ligandList)
			
		return conformations

	def _replaceMetal(self, combo, conf_list):
		for elem in conf_list:
			if 'X' in elem:
				elem[0] = combo.metal.name
				break
		return conf_list

	def _transformSameLigs(self, coords, ligand_list):
		conformation = []
		conformation.append(list(coords[0]))
		for i, lig in enumerate(ligand_list):
			coord = np.array(coords[i+1][2:], dtype=np.float32)
			ligCoords = [atom[1:] for atom in lig.atom_list]
			ligCoords = np.array(ligCoords)
			ligCoords = coord + ligCoords
			newLig = self._prependAtomInfo(lig, ligCoords)
			for each in newLig:
				conformation.append(each)

		return conformation

	def _prependAtomInfo(self, ligand, coords):
		atoms = []
		for atom, coord in zip(ligand.atom_list, coords.tolist()):
			newAtom = []
			newAtom.append(str(atom[0]))
			newAtom.append('0')
			for i in range(3):
				newAtom.append(str(coord[i]))
				
			atoms.append(newAtom)
		
		return atoms
			
	
	def _transformMixedLigs(self):
		print("I'm not defined yet!")
		sys.exit()



from setuptools import setup, find_packages

from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get description
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
	long_description = f.read()

setup(
	name='zeoliteclusterizer',
	version='0.1.1',
	description='A python module built for rapid computational screening of catalysts on rigid structures',
	url='https://github.com/ngallup/zeoliteclusterizer',

	author='Nathan Gallup',
	author_email='gallup@chem.ucla.edu',

	license='Free for non-commercial use',

	classifiers=[
		'Development Status :: 3 - Alpha',
		'Intended Audience :: Developers',
		'Intended Audience :: Science/Research',
		'Topic :: Scientific/Engineering',
		'Topic :: Scientific/Engineering :: Chemistry',
		'License :: Free for non-commercial use',
		'Programming Language :: Python :: 2',
		'Programming Language :: Python :: 3'
	],

	keywords='chemistry quantum mechanics zeolites mofs catalysis gaussian',

	packages=find_packages(exclude=['*tests/tests/*']),
)
